from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any
from typing import Dict
from typing import Literal
from typing import NewType
from typing import Tuple
from uuid import UUID

import dateutil.parser

from kleinkram.errors import ParsingError
from kleinkram.models import File
from kleinkram.models import FileState
from kleinkram.models import Mission
from kleinkram.models import Project

__all__ = [
    "_parse_project",
    "_parse_mission",
    "_parse_file",
]


ProjectObject = NewType("ProjectObject", Dict[str, Any])
MissionObject = NewType("MissionObject", Dict[str, Any])
FileObject = NewType("FileObject", Dict[str, Any])

MISSION = "mission"
PROJECT = "project"


class FileObjectKeys(str, Enum):
    UUID = "uuid"
    FILENAME = "filename"
    DATE = "date"  # at some point this will become a metadata
    CREATED_AT = "createdAt"
    UPDATED_AT = "updatedAt"
    STATE = "state"
    SIZE = "size"
    HASH = "hash"
    TYPE = "type"
    CATEGORIES = "categories"


class MissionObjectKeys(str, Enum):
    UUID = "uuid"
    NAME = "name"
    DESCRIPTION = "description"
    CREATED_AT = "createdAt"
    UPDATED_AT = "updatedAt"


class ProjectObjectKeys(str, Enum):
    UUID = "uuid"
    NAME = "name"
    DESCRIPTION = "description"
    CREATED_AT = "createdAt"
    UPDATED_AT = "updatedAt"


def _get_nested_info(data, key: Literal["mission", "project"]) -> Tuple[UUID, str]:
    nested_data = data[key]
    return (
        UUID(nested_data[ProjectObjectKeys.UUID], version=4),
        nested_data[ProjectObjectKeys.NAME],
    )


def _parse_datetime(date: str) -> datetime:
    try:
        return dateutil.parser.isoparse(date)
    except ValueError as e:
        raise ParsingError(f"error parsing date: {date}") from e


def _parse_file_state(state: str) -> FileState:
    try:
        return FileState(state)
    except ValueError as e:
        raise ParsingError(f"error parsing file state: {state}") from e


def _parse_project(project_object: ProjectObject) -> Project:
    try:
        id_ = UUID(project_object[ProjectObjectKeys.UUID], version=4)
        name = project_object[ProjectObjectKeys.NAME]
        description = project_object[ProjectObjectKeys.DESCRIPTION]
        created_at = _parse_datetime(project_object[ProjectObjectKeys.CREATED_AT])
        updated_at = _parse_datetime(project_object[ProjectObjectKeys.UPDATED_AT])
    except Exception as e:
        raise ParsingError(f"error parsing project: {project_object}") from e
    return Project(
        id=id_,
        name=name,
        description=description,
        created_at=created_at,
        updated_at=updated_at,
    )


def _parse_mission(mission: MissionObject) -> Mission:
    try:
        id_ = UUID(mission[MissionObjectKeys.UUID], version=4)
        name = mission[MissionObjectKeys.NAME]
        created_at = _parse_datetime(mission[MissionObjectKeys.CREATED_AT])
        updated_at = _parse_datetime(mission[MissionObjectKeys.UPDATED_AT])
        metadata = {}  # TODO: this crap is really bad to parse

        project_id, project_name = _get_nested_info(mission, PROJECT)

        parsed = Mission(
            id=id_,
            name=name,
            created_at=created_at,
            updated_at=updated_at,
            metadata=metadata,
            project_id=project_id,
            project_name=project_name,
        )
    except Exception as e:
        raise ParsingError(f"error parsing mission: {mission}") from e
    return parsed


def _parse_file(file: FileObject) -> File:
    try:
        name = file[FileObjectKeys.FILENAME]
        id_ = UUID(file[FileObjectKeys.UUID], version=4)
        fsize = file[FileObjectKeys.SIZE]
        fhash = file[FileObjectKeys.HASH]
        ftype = file[FileObjectKeys.TYPE].split(".")[-1]
        fdate = file[FileObjectKeys.DATE]
        created_at = _parse_datetime(file[FileObjectKeys.CREATED_AT])
        updated_at = _parse_datetime(file[FileObjectKeys.UPDATED_AT])
        state = _parse_file_state(file[FileObjectKeys.STATE])
        categories = file[FileObjectKeys.CATEGORIES]

        mission_id, mission_name = _get_nested_info(file, MISSION)
        project_id, project_name = _get_nested_info(file[MISSION], PROJECT)

        parsed = File(
            id=id_,
            name=name,
            hash=fhash,
            size=fsize,
            type_=ftype,
            date=fdate,
            categories=categories,
            state=state,
            created_at=created_at,
            updated_at=updated_at,
            mission_id=mission_id,
            mission_name=mission_name,
            project_id=project_id,
            project_name=project_name,
        )
    except Exception as e:
        raise ParsingError(f"error parsing file: {file}") from e
    return parsed



from __future__ import annotations

import urllib.parse
import webbrowser
from getpass import getpass
from http.server import BaseHTTPRequestHandler
from http.server import HTTPServer
from typing import Optional

from kleinkram.config import CONFIG_PATH
from kleinkram.config import Credentials
from kleinkram.config import get_config
from kleinkram.config import save_config

CLI_CALLBACK_ENDPOINT = "/cli/callback"
OAUTH_SLUG = "/auth/"


def _has_browser() -> bool:
    try:
        webbrowser.get()
        return True
    except webbrowser.Error:
        return False


def _headless_auth(*, url: str) -> None:

    print(f"please open the following URL manually to authenticate: {url}")
    print("enter the authentication token provided after logging in:")
    auth_token = getpass("authentication token: ")
    refresh_token = getpass("refresh token: ")

    if auth_token and refresh_token:
        config = get_config()
        config.credentials = Credentials(
            auth_token=auth_token, refresh_token=refresh_token
        )
        save_config(config)
        print(f"Authentication complete. Tokens saved to {CONFIG_PATH}.")
    else:
        raise ValueError("Please provided tokens.")


class OAuthCallbackHandler(BaseHTTPRequestHandler):
    def do_GET(self):
        if self.path.startswith(CLI_CALLBACK_ENDPOINT):
            query = urllib.parse.urlparse(self.path).query
            params = urllib.parse.parse_qs(query)

            try:
                creds = Credentials(
                    auth_token=params.get("authtoken")[0],  # type: ignore
                    refresh_token=params.get("refreshtoken")[0],  # type: ignore
                )
                config = get_config()
                config.credentials = creds
                save_config(config)
            except Exception:
                raise RuntimeError("Failed to fetch authentication tokens.")

            self.send_response(200)
            self.send_header("Content-type", "text/html")
            self.end_headers()
            self.wfile.write(b"Authentication successful. You can close this window.")
        else:
            raise RuntimeError("Invalid path")

    def log_message(self, *args, **kwargs):
        _ = args, kwargs
        pass  # suppress logging


def _browser_auth(*, url: str) -> None:
    webbrowser.open(url)

    server = HTTPServer(("", 8000), OAuthCallbackHandler)
    server.handle_request()

    print(f"Authentication complete. Tokens saved to {CONFIG_PATH}.")


def login_flow(
    *, oAuthProvider: str, key: Optional[str] = None, headless: bool = False
) -> None:
    config = get_config()
    # use cli key login
    if key is not None:
        config.credentials = Credentials(api_key=key)
        save_config(config)
        return

    oauth_url = f"{config.endpoint.api}{OAUTH_SLUG}{oAuthProvider}?state=cli"
    if not headless and _has_browser():
        _browser_auth(url=oauth_url)
    else:
        _headless_auth(url=f"{oauth_url}-no-redirect")



from __future__ import annotations

import sys
from collections import OrderedDict
from typing import Any
from typing import Callable
from typing import Type

import typer
from click import ClickException
from rich.console import Console
from rich.panel import Panel

from kleinkram.utils import upper_camel_case_to_words

ExceptionHandler = Callable[[Exception], int]


class ErrorHandledTyper(typer.Typer):
    """\
    error handlers that are last added will be used first
    """

    _error_handlers: OrderedDict[Type[Exception], ExceptionHandler]

    def error_handler(
        self, exc: Type[Exception]
    ) -> Callable[[ExceptionHandler], ExceptionHandler]:
        def dec(func: ExceptionHandler) -> ExceptionHandler:
            self._error_handlers[exc] = func
            return func

        return dec

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._error_handlers = OrderedDict()

    def __call__(self, *args: Any, **kwargs: Any) -> int:
        try:
            return super().__call__(*args, **kwargs)
        except Exception as e:
            if isinstance(e, ClickException):
                raise
            for tp, handler in reversed(self._error_handlers.items()):
                if isinstance(e, tp):
                    exit_code = handler(e)
                    raise SystemExit(exit_code)
            raise


def display_error(*, exc: Exception, verbose: bool) -> None:
    split_exc_name = upper_camel_case_to_words(type(exc).__name__)

    if verbose:
        panel = Panel(
            str(exc),  # get the error message
            title=" ".join(split_exc_name),
            style="red",
            border_style="bold",
        )
        Console(file=sys.stderr).print(panel)
    else:
        text = f"{type(exc).__name__}"
        if str(exc):
            text += f": {exc}"
        print(text, file=sys.stderr)



from __future__ import annotations

from typing import Optional

import typer

import kleinkram.api.routes
import kleinkram.core
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import ProjectQuery
from kleinkram.api.routes import get_project
from kleinkram.config import get_shared_state
from kleinkram.printing import print_project_info
from kleinkram.utils import split_args

project_typer = typer.Typer(
    no_args_is_help=True, context_settings={"help_option_names": ["-h", "--help"]}
)


NOT_IMPLEMENTED_YET = """\
Not implemented yet, open an issue if you want specific functionality
"""

CREATE_HELP = "create a project"
INFO_HELP = "get information about a project"
UPDATE_HELP = "update a project"
DELETE_HELP = "delete a project"


@project_typer.command(help=CREATE_HELP)
def create(
    project: str = typer.Option(..., "--project", "-p", help="project name"),
    description: str = typer.Option(
        ..., "--description", "-d", help="project description"
    ),
) -> None:
    client = AuthenticatedClient()
    project_id = kleinkram.api.routes._create_project(client, project, description)

    project_parsed = get_project(client, ProjectQuery(ids=[project_id]))
    print_project_info(project_parsed, pprint=get_shared_state().verbose)


@project_typer.command(help=INFO_HELP)
def info(
    project: str = typer.Option(..., "--project", "-p", help="project id or name")
) -> None:
    project_ids, project_patterns = split_args([project])
    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)

    client = AuthenticatedClient()
    project_parsed = get_project(client=client, query=project_query)
    print_project_info(project_parsed, pprint=get_shared_state().verbose)


@project_typer.command(help=UPDATE_HELP)
def update(
    project: str = typer.Option(..., "--project", "-p", help="project id or name"),
    description: Optional[str] = typer.Option(
        None, "--description", "-d", help="project description"
    ),
    new_name: Optional[str] = typer.Option(
        None, "--new-name", "-n", "--name", help="new project name"
    ),
) -> None:
    if description is None and new_name is None:
        raise typer.BadParameter(
            "nothing to update, provide --description or --new-name"
        )

    project_ids, project_patterns = split_args([project])
    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)

    client = AuthenticatedClient()
    project_id = get_project(client=client, query=project_query).id
    kleinkram.core.update_project(
        client=client, project_id=project_id, description=description, new_name=new_name
    )

    project_parsed = get_project(client, ProjectQuery(ids=[project_id]))
    print_project_info(project_parsed, pprint=get_shared_state().verbose)


@project_typer.command(help=DELETE_HELP)
def delete(
    project: str = typer.Option(..., "--project", "-p", help="project id or name")
) -> None:
    project_ids, project_patterns = split_args([project])
    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)

    client = AuthenticatedClient()
    project_id = get_project(client=client, query=project_query).id
    kleinkram.core.delete_project(client=client, project_id=project_id)


@project_typer.command(help=NOT_IMPLEMENTED_YET)
def prune() -> None:
    raise NotImplementedError(NOT_IMPLEMENTED_YET)



from __future__ import annotations

from uuid import uuid4

import pytest

from kleinkram.api.query import InvalidMissionQuery
from kleinkram.api.query import InvalidProjectQuery
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.api.query import check_mission_query_is_creatable
from kleinkram.api.query import check_project_query_is_creatable
from kleinkram.api.query import mission_query_is_unique
from kleinkram.api.query import project_query_is_unique


@pytest.mark.parametrize(
    "query, expected",
    [
        pytest.param(MissionQuery(), False, id="match all"),
        pytest.param(MissionQuery(patterns=["*"]), False, id="mission name match all"),
        pytest.param(
            MissionQuery(patterns=["test"]),
            False,
            id="mission name without project",
        ),
        pytest.param(
            MissionQuery(patterns=["test"], project_query=ProjectQuery()),
            False,
            id="mission name with non-unique project",
        ),
        pytest.param(
            MissionQuery(
                patterns=["test"],
                project_query=ProjectQuery(ids=[uuid4()]),
            ),
            True,
            id="mission name with unique project",
        ),
        pytest.param(
            MissionQuery(ids=[uuid4()]),
            True,
            id="mission by id",
        ),
        pytest.param(
            MissionQuery(ids=[uuid4(), uuid4()]),
            False,
            id="multiple mission ids",
        ),
    ],
)
def test_mission_query_is_unique(query, expected):
    assert mission_query_is_unique(query) == expected


@pytest.mark.parametrize(
    "query, expected",
    [
        pytest.param(ProjectQuery(), False, id="match all"),
        pytest.param(ProjectQuery(patterns=["*"]), False, id="project name match all"),
        pytest.param(
            ProjectQuery(patterns=["test"]),
            True,
            id="project name",
        ),
        pytest.param(
            ProjectQuery(ids=[uuid4()]),
            True,
            id="project by id",
        ),
        pytest.param(
            ProjectQuery(ids=[uuid4(), uuid4()]),
            False,
            id="multiple project ids",
        ),
    ],
)
def test_project_query_is_unique(query, expected):
    assert project_query_is_unique(query) == expected


@pytest.mark.parametrize(
    "query, valid",
    [
        pytest.param(
            MissionQuery(patterns=["test"], project_query=ProjectQuery()),
            False,
            id="non-unique project",
        ),
        pytest.param(
            MissionQuery(
                patterns=["test"],
                project_query=ProjectQuery(ids=[uuid4()]),
            ),
            True,
            id="valid query",
        ),
        pytest.param(
            MissionQuery(ids=[uuid4()]),
            False,
            id="mission by id",
        ),
    ],
)
def test_check_mission_query_is_createable(query, valid):
    if not valid:
        with pytest.raises(InvalidMissionQuery):
            check_mission_query_is_creatable(query)
    else:
        check_mission_query_is_creatable(query)


@pytest.mark.parametrize(
    "query, valid",
    [
        pytest.param(
            ProjectQuery(patterns=["test"]),
            True,
            id="project name",
        ),
        pytest.param(
            ProjectQuery(ids=[uuid4()]),
            False,
            id="project by id",
        ),
        pytest.param(
            ProjectQuery(ids=[uuid4(), uuid4()]),
            False,
            id="multiple project ids",
        ),
    ],
)
def test_check_project_query_is_creatable(query, valid):
    if not valid:
        with pytest.raises(InvalidProjectQuery):
            check_project_query_is_creatable(query)
    else:
        check_project_query_is_creatable(query)



from __future__ import annotations

import logging
from pathlib import Path
from typing import List
from typing import Optional

import typer

import kleinkram.core
from kleinkram.api.client import AuthenticatedClient
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.config import get_shared_state
from kleinkram.printing import print_file_verification_status
from kleinkram.utils import split_args

logger = logging.getLogger(__name__)


HELP = """\
Verify if files were uploaded correctly.
"""

verify_typer = typer.Typer(name="verify", invoke_without_command=True, help=HELP)


@verify_typer.callback()
def verify(
    files: List[str] = typer.Argument(help="files to upload"),
    project: Optional[str] = typer.Option(
        None, "--project", "-p", help="project id or name"
    ),
    mission: str = typer.Option(..., "--mission", "-m", help="mission id or name"),
    skip_hash: bool = typer.Option(None, help="skip hash check"),
    check_file_hash: bool = typer.Option(
        True,
        help="check file hash. If True, file names and file hashes are checked.",
    ),
    check_file_size: bool = typer.Option(
        True,
        help="check file size. If True, file names and file sizes are checked.",
    ),
) -> None:
    # get all filepaths
    file_paths = [Path(file) for file in files]

    # get mission query
    mission_ids, mission_patterns = split_args([mission])
    project_ids, project_patterns = split_args([project] if project else [])
    project_query = ProjectQuery(ids=project_ids, patterns=project_patterns)
    mission_query = MissionQuery(
        ids=mission_ids, patterns=mission_patterns, project_query=project_query
    )

    verbose = get_shared_state().verbose
    file_status = kleinkram.core.verify(
        client=AuthenticatedClient(),
        query=mission_query,
        file_paths=file_paths,
        skip_hash=skip_hash,
        check_file_hash=check_file_hash,
        check_file_size=check_file_size,
        verbose=verbose,
    )
    print_file_verification_status(file_status, pprint=verbose)



from __future__ import annotations

from uuid import uuid4

from kleinkram.api.query import FileQuery
from kleinkram.api.query import MissionQuery
from kleinkram.api.query import ProjectQuery
from kleinkram.wrappers import _args_to_file_query
from kleinkram.wrappers import _args_to_mission_query
from kleinkram.wrappers import _args_to_project_query


def test_args_to_project_query() -> None:
    assert _args_to_project_query() == ProjectQuery()
    assert _args_to_project_query(project_names=["test"]) == ProjectQuery(
        patterns=["test"]
    )

    _id = uuid4()
    assert _args_to_project_query(project_ids=[_id]) == ProjectQuery(ids=[_id])
    assert _args_to_project_query(
        project_names=["test"], project_ids=[_id]
    ) == ProjectQuery(patterns=["test"], ids=[_id])
    assert _args_to_project_query(project_ids=[str(_id)]) == ProjectQuery(ids=[_id])


def test_args_to_mission_query() -> None:
    assert _args_to_mission_query() == MissionQuery()
    assert _args_to_mission_query(mission_names=["test"]) == MissionQuery(
        patterns=["test"]
    )

    _id = uuid4()
    assert _args_to_mission_query(mission_ids=[_id]) == MissionQuery(ids=[_id])
    assert _args_to_mission_query(
        mission_names=["test"], mission_ids=[_id]
    ) == MissionQuery(patterns=["test"], ids=[_id])
    assert _args_to_mission_query(mission_ids=[str(_id)]) == MissionQuery(ids=[_id])

    assert _args_to_mission_query(project_names=["test"]) == MissionQuery(
        project_query=ProjectQuery(patterns=["test"])
    )
    assert _args_to_mission_query(project_ids=[_id]) == MissionQuery(
        project_query=ProjectQuery(ids=[_id])
    )


def test_args_to_file_query() -> None:
    assert _args_to_file_query() == FileQuery()
    assert _args_to_file_query(file_names=["test"]) == FileQuery(patterns=["test"])

    _id = uuid4()
    assert _args_to_file_query(file_ids=[_id]) == FileQuery(ids=[_id])
    assert _args_to_file_query(file_names=["test"], file_ids=[_id]) == FileQuery(
        patterns=["test"], ids=[_id]
    )
    assert _args_to_file_query(file_ids=[str(_id)]) == FileQuery(ids=[_id])

    assert _args_to_file_query(mission_names=["test"]) == FileQuery(
        mission_query=MissionQuery(patterns=["test"])
    )
    assert _args_to_file_query(mission_ids=[_id]) == FileQuery(
        mission_query=MissionQuery(ids=[_id])
    )

    assert _args_to_file_query(project_names=["test"]) == FileQuery(
        mission_query=MissionQuery(project_query=ProjectQuery(patterns=["test"]))
    )
    assert _args_to_file_query(project_ids=[_id]) == FileQuery(
        mission_query=MissionQuery(project_query=ProjectQuery(ids=[_id]))
    )
